/** Simple implementation of a fraction algebra (example of a <i>user defined algebra</i>). <br>
  * Stores numerator and denominator with int precision. Allows mixed operations with int and double.<br>
  * For a may be more elaborate implementation 
  * see &nbsp;<a target="_top" href="../programmer_examples/commons_math_examples/0_CommonsMath_README.html">Commons Math</a> &rarr;
  * <a target="_blank" href="http://commons.apache.org/proper/commons-math/javadocs/api-3.6.1/index.html?org/apache/commons/math3/fraction/BigFraction.html"><code>BigFraction</code></a>
  * <!-- <a target="_blank" href="http://commons.apache.org/proper/commons-math/apidocs/index.html?org/apache/commons/math4/fraction/BigFraction.html"><code>BigFraction</code></a>.-->
  * @see <a target="_top" href="../programmer_examples/MyFractionAlgebra_UserDefinedAlgebra.html">user defined algebra</a>
  */
public class MyFractionAlgebra implements JxnCloneableAlgebra
{
   static final boolean DEBUG = false;
   
   private int numerator;
   private int denominator;

   /** Constructs a new Fraction object. */
   public MyFractionAlgebra( int numerator, int denominator )
   {
      this.numerator = numerator;
      this.denominator = denominator;
      reduce();
   }

   /** Equivalent to {@code new MyFractionAlgebra( numerator, 1 )}. <br>
     * Required for mixed operations: int <i>op</i> fraction, fraction <i>op</i> int
     */
   public MyFractionAlgebra( int numerator )
   {
      this( numerator, 1 );
   }

   /** Finds a fraction that represents {@code x}, using the continued fraction algorithm. <br>
     * Required for mixed operations: double <i>op</i> fraction, fraction <i>op</i> double.
     * @see <a target="_blank" href="https://en.wikipedia.org/wiki/Continued_fraction">continued fraction</a>
     */
   public MyFractionAlgebra( double x )
   {
      if( x == 0 )
      {
         numerator = 0;
         denominator  = 1;
         return;
      }

      // Kettenbruchalgorithmus (continued fraction)
      int zPrev = 0;
      int nPrev = 1;
      double ax = Math.abs(x);
      numerator = 1;
      denominator = 0;
      
      if( DEBUG ) System.out.print( x );

      while( true )
      {
         if( ax > Integer.MAX_VALUE ) break;
         int ai = (int)Math.floor( ax );
         if( DEBUG ) System.out.print( " " + ai );
         
         long zNext = ai * (long)numerator;
         long nNext = ai * (long)denominator;
         if( zNext > Integer.MAX_VALUE ) break;
         if( nNext > Integer.MAX_VALUE ) break;

         zNext += zPrev;
         nNext += nPrev;
         if( zNext > Integer.MAX_VALUE ) break;
         if( nNext > Integer.MAX_VALUE ) break;

         zPrev = numerator;
         nPrev = denominator;
         numerator = (int)zNext;
         denominator  = (int)nNext;

         ax %= 1.;
         if( ax == 0. ) break;
         ax = 1. / ax;
      }
      
      if( x < 0. ) numerator = -numerator;
      reduce();
      if( DEBUG ) System.out.println( " => " + this );

      if( Math.abs( x - doubleValue() ) > Math.abs(x) * 2./Integer.MAX_VALUE )
      {
         System.out.println( "??? (3) " + this + " <> " + x );
         throw new KmgFormelException( "unable to convert " + x + " to " + getClass() );
      }
   }

   /** Returns the internally stored numerator. */
   public int getNumerator()
   {
      return numerator;
   }
   
   /** Returns the internally stored denominator. */
   public int getDenominator()
   {
      return denominator;
   }

   /** Returns this fraction as double. */
   public double doubleValue()
   {
      return (double)numerator / denominator;
   }


// --- interface methods ---

   /** <b>Note:</b> {@code MyFractionAlgebra} implements {@link JxnCloneableAlgebra} only for demonstration.
     * As {@code MyFractionAlgebra} has only 2 int members {@code numerator} and {@code denominator} it
     * should better be implemented as <i>immutable</i> like {@code java.math.}{@link java.math.BigInteger} 
     * or <!-- <a target="_blank" href="http://commons.apache.org/proper/commons-math/userguide/fraction.html">CommonsMath</a>  -->
     * <a target="_top" href="../programmer_examples/commons_math_examples/0_CommonsMath_README.html">Commons Math</a> &rarr;
     * <a target="_blank" href="http://commons.apache.org/proper/commons-math/javadocs/api-3.6.1/index.html?org/apache/commons/math3/fraction/BigFraction.html"><code>BigFraction</code></a>
     * <!-- <a href="http://commons.apache.org/proper/commons-math/apidocs/org/apache/commons/math4/fraction/BigFraction.html"><code>BigFraction</code></a> -->.<br>
     * <br>
     * <b>Description copied from interface: {@link JxnCloneableAlgebra}</b><br>
     * {@inheritDoc}
     */
   public JxnCloneableAlgebra cloneThis()
   {
      return new MyFractionAlgebra( numerator, denominator );
   }


// --- reflection methods ---

   /** Adds {@code opnd} to this.
     * @return this
     */
   public MyFractionAlgebra add( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator + denominator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      if( DEBUG ) System.out.println( "add(0): " + this + " " + opnd );
      long g = gcd( denominator, opnd.denominator );
      long n1 = opnd.denominator / g * numerator;
      long n2 = denominator / g * opnd.numerator;
      if( DEBUG ) System.out.println( "add(1): " + g + " " + n1 + " " + n2 );
      if( Math.abs(n1) > Integer.MAX_VALUE || Math.abs(n2) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (+) " + opnd );
      }
      long num = n1 + n2;
      long den = denominator / g * opnd.denominator;
      if( DEBUG ) System.out.println( "add(2): " + num + " " + den );
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (+) " + opnd );
      }
      numerator = (int)num;
      denominator  = (int)den;
      reduce();
      return this;
   }
   
/* Subtracts opnd from this (Implementation not necesary, if constructor MyFractionAlgebra(int) exists).
   public MyFractionAlgebra sub( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator - denominator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      long g = gcd( denominator, opnd.denominator );
      long n1 = opnd.denominator / g * numerator;
      long n2 = denominator / g * opnd.numerator;
      if( Math.abs(n1) > Integer.MAX_VALUE || Math.abs(n2) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (-) " + opnd );
      }
      long num = n1 - n2;
      long den = denominator / g * opnd.denominator;
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (-) " + opnd );
      }
      numerator = (int)num;
      denominator  = (int)den;
      reduce();
      return this;
   }
*/
   
   /** (Post)Multiplies this by {@code opnd}.
     * @return this
     */
   public MyFractionAlgebra mul( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).numerator;
   // denominator= denominator*((MyFractionAlgebra)opnd).denominator;
      long g1 = gcd( numerator, opnd.denominator );
      long g2 = gcd( denominator,  opnd.numerator );
      long num  = numerator / g1 * ( opnd.numerator / g2 );
      long den  = denominator  / g2 * ( opnd.denominator  / g1 );
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (*) " + opnd );
      }
      numerator = (int)num;
      denominator  = (int)den;
      reduce();
      return this;
   }

/* Divides this by opnd (Implementation not necessary, if method inv() exists).
   public MyFractionAlgebra div( MyFractionAlgebra opnd )
   {
   // numerator = numerator*((MyFractionAlgebra)opnd).denominator;
   // denominator = denominator*((MyFractionAlgebra)opnd).numerator;
      long g1 = gcd( numerator, ((MyFractionAlgebra)opnd).numerator );
      long g2 = gcd( denominator,  ((MyFractionAlgebra)opnd).denominator );
      long num  = numerator / g1 * ( ((MyFractionAlgebra)opnd).denominator  / g2 );
      long den  = denominator  / g2 * ( ((MyFractionAlgebra)opnd).numerator / g1 );
      if( Math.abs(num) > Integer.MAX_VALUE || Math.abs(den) > Integer.MAX_VALUE )
      {
         throw new KmgFormelException( "integer overflow for: " + this + " (/) " + opnd );
      }
      numerator = (int)num;
      denominator  = (int)den;
      reduce();
      return this;
   }
*/

   /** Replaces this by its reciprocal.
     * @return this
     */
   public MyFractionAlgebra inv()
   {
      int temp = numerator;
      numerator = denominator;
      denominator = temp;
      return this;
   }
   
   /** Replaces this by its power of {@code exponent}.
     * @return this
     */
   public MyFractionAlgebra pow( int exponent )
   {
      if( exponent < 0 )
      {
         int tmp = numerator;
         numerator   = power( denominator, -exponent );
         denominator = power( tmp, -exponent );
      }
      else
      {
         numerator   = power( numerator, exponent );
         denominator = power( denominator, exponent );
      }
   // reduce();  // not neccessary
      return this;
   }


// --- overwriten methods ---

   /** Fractions are equal, if both numerator and denominator are equal. */
// public boolean equals( MyFractionAlgebra other )  // geht nicht
   public boolean equals( Object other )
   {
      if( numerator != ((MyFractionAlgebra)other).numerator ) return false;
      if( denominator != ((MyFractionAlgebra)other).denominator ) return false;
      return true;
   }
   
/* TODO: equals => gleicher Hashcode
   public int hashCode()
   {
      return ...
   }
*/

   /** Returns a string representation of the fraction. */
   public String toString()
   {
      return numerator + " / " + denominator + " = " + doubleValue();
   }
   
   
// --- other reflection methods ---

   /** Returns the greatest common divisor (gcd) of {@code m} and {@code n}. */
   public static int gcd( int m, int n )
   {
//    if( m > n )
//       return gcd( m - n, n );
//    else if( n > m )
//       return gcd( m, n - m );
//    else
//       return m;

      if( n == 0 ) return 0;

      for(;;)
      {
         int mn = m % n;
         if( mn == 0 ) return Math.abs(n);
         m = n;
         n = mn;
      }
   }
   
   
// --- internal methods ---

   /** Reduces the fraction: devides numerator and denominator by their greatest common divisor. */
   protected void reduce()  // reduce
   {
      int g = gcd( numerator, denominator );
      if ( g == 0 ) return;
      if( denominator < 0 ) g = -g;
      numerator /= g;
      denominator /= g;
   }
   
   /** power.
     * @param exponent must be &gt;= 0
     */
   protected int power( int base, int exponent )
   {
      if( exponent < 0 ) throw new KmgFormelException( "negative exponent not allowed" );
      int exp = exponent;
      
      long result = 1;
      while( exp > 0 )
      {
         result *= base;
         if( Math.abs( result ) > Integer.MAX_VALUE ) throw new KmgFormelException( "integer overflow for: " + this + " ^ " + exponent );
         exp--;
      }
      
      return (int)result;
   }
}
